/* 
 * Copyright (C) 2013 Lars Marowsky-Bree <lmb@suse.com>
 * 
 * Based on crm_mon.c, which was:
 * Copyright (C) 2004 Andrew Beekhof <andrew@beekhof.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <config.h>
#include <crm_config.h>

#include <crm/cluster.h>
#include <crm/common/mainloop.h>

#if CHECK_TWO_NODE
#include <glib-unix.h>
#endif

#if SUPPORT_COROSYNC
#include <corosync/quorum.h>
#include <corosync/cmap.h>
#include <gio/gio.h>
#include <time.h>
#endif

#include "sbd.h"

//undef SUPPORT_PLUGIN
//define SUPPORT_PLUGIN 1

static bool remote_node = false;
static pid_t remoted_pid = 0;
static int reconnect_msec = 1000;
static GMainLoop *mainloop = NULL;
static guint notify_timer = 0;
static crm_cluster_t cluster;
static gboolean sbd_remote_check(gpointer user_data);
static long unsigned int find_pacemaker_remote(void);
static void sbd_membership_destroy(gpointer user_data);


#if SUPPORT_PLUGIN
static void
sbd_plugin_membership_dispatch(cpg_handle_t handle,
                           const struct cpg_name *groupName,
                           uint32_t nodeid, uint32_t pid, void *msg, size_t msg_len)
{
    if(msg_len > 0) {
        set_servant_health(pcmk_health_online, LOG_INFO,
                           "Connected to %s", name_for_cluster_type(get_cluster_type()));
    } else {
        set_servant_health(pcmk_health_unclean, LOG_WARNING,
                           "Broken %s message", name_for_cluster_type(get_cluster_type()));
    }
    notify_parent();
    return;
}
#endif /* SUPPORT_PLUGIN */

#if SUPPORT_COROSYNC

static bool two_node = false;
static bool ever_seen_both = false;
static int cpg_membership_entries = -1;
static cmap_handle_t cmap_handle = 0;
static GSource *cmap_source = NULL;

static cmap_track_handle_t twonode_track_handle = 0;
static cmap_track_handle_t consensus_track_handle = 0;

struct qdata {
    quorum_handle_t q_handle;
    uint32_t        q_type;
    int             running;
    int             consensus_timeout;
    bool            ever_had_quorum;
    struct timespec ts;
};
static struct qdata qd;

void
sbd_cpg_membership_health_update()
{
    if(cpg_membership_entries > 0) {
        bool quorum_is_suspect =
            (two_node && ever_seen_both && cpg_membership_entries == 1);

        if (!quorum_is_suspect) {
            set_servant_health(pcmk_health_online, LOG_INFO,
                           "Connected to %s (%u members)",
                           name_for_cluster_type(get_cluster_type()),
                           cpg_membership_entries
                          );
        } else {
            /* Alternative would be asking votequorum for number of votes.
             * Using pacemaker's cpg as source for number of active nodes
             * avoids binding to an additional library, is definitely
             * less code to write and we wouldn't have to combine data
             * from 3 sources (cmap, cpq & votequorum) in a potentially
             * racy environment.
             */
            set_servant_health(pcmk_health_noquorum, LOG_WARNING,
                           "Connected to %s but requires both nodes present",
                           name_for_cluster_type(get_cluster_type())
                          );
        }

        if (cpg_membership_entries > 1) {
            ever_seen_both = true;
        }
    } else {
        set_servant_health(pcmk_health_unclean, LOG_WARNING,
                           "Empty %s membership", name_for_cluster_type(get_cluster_type()));
    }
}

void
sbd_cpg_membership_dispatch(cpg_handle_t handle,
                    const struct cpg_name *groupName,
                    const struct cpg_address *member_list, size_t member_list_entries,
                    const struct cpg_address *left_list, size_t left_list_entries,
                    const struct cpg_address *joined_list, size_t joined_list_entries)
{
    cpg_membership_entries = member_list_entries;
    sbd_cpg_membership_health_update();
    notify_parent();
}

static gboolean
cmap_dispatch_callback (gpointer user_data)
{
    cmap_dispatch(cmap_handle, CS_DISPATCH_ALL);
    return TRUE;
}

#if CHECK_TWO_NODE
static void sbd_cmap_notify_fn(
    cmap_handle_t cmap_handle,
    cmap_track_handle_t cmap_track_handle,
    int32_t event,
    const char *key_name,
    struct cmap_notify_value new_val,
    struct cmap_notify_value old_val,
    void *user_data)
{
    if (new_val.type == CMAP_VALUETYPE_UINT8) {
        switch (event) {
            case CMAP_TRACK_ADD:
            case CMAP_TRACK_MODIFY:
                two_node = *((uint8_t *) new_val.data);
                break;
            case CMAP_TRACK_DELETE:
                two_node = false;
                break;
            default:
                return;
        }
        sbd_cpg_membership_health_update();
        notify_parent();
    }
}
#endif /* CHECK_TWO_NODE */

static void cmap_consensus_notify_fn(cmap_handle_t cmap_handle,
                                     cmap_track_handle_t cmap_track_handle,
                                     int32_t event,
                                     const char *key_name,
                                     struct cmap_notify_value new_val,
                                     struct cmap_notify_value old_val,
                                     void *user_data)
{
    if (new_val.type == CMAP_VALUETYPE_UINT32) {
        switch (event) {
            case CMAP_TRACK_ADD:
            case CMAP_TRACK_MODIFY:
                qd.consensus_timeout = *((uint32_t *) new_val.data);
                /* Take 1s (in milliseconds) off the timeout because we only
                 * check once per second */
                if (qd.consensus_timeout > 2000) qd.consensus_timeout -= 1000;
                break;
            case CMAP_TRACK_DELETE:
                qd.consensus_timeout = 0; /* This should never happen though */
                break;
            default:
                return;
        }
    }
}

static gboolean
sbd_crm_setup(void)
{
    uint8_t two_node_u8 = 0;
    uint32_t consensus = 0;
    int cmap_fd;

    if (!cmap_handle) {
        if (cmap_initialize(&cmap_handle) != CS_OK) {
            cl_log(LOG_WARNING, "Cannot initialize CMAP service\n");
            goto out;
        }

        if (cmap_track_add(cmap_handle, "runtime.config.totem.consensus",
                            CMAP_TRACK_DELETE|CMAP_TRACK_MODIFY|CMAP_TRACK_ADD,
                            cmap_consensus_notify_fn, NULL, &consensus_track_handle) != CS_OK) {
            cl_log(LOG_WARNING, "Failed adding CMAP tracker for runtime.config.totem.consensus\n");
            goto out;
        }
#if CHECK_TWO_NODE
        if (cmap_track_add(cmap_handle, "quorum.two_node",
                            CMAP_TRACK_DELETE|CMAP_TRACK_MODIFY|CMAP_TRACK_ADD,
                            sbd_cmap_notify_fn, NULL, &twonode_track_handle) != CS_OK) {
            cl_log(LOG_WARNING, "Failed adding CMAP tracker for 2Node-mode\n");
            goto out;
        }
#endif /* CHECK_TWO_NODE */

        /* add the tracker to mainloop */
        if (cmap_fd_get(cmap_handle, &cmap_fd) != CS_OK) {
            cl_log(LOG_WARNING, "Failed to get a file handle for cmap\n");
            goto out;
        }

        if (!(cmap_source = g_unix_fd_source_new (cmap_fd, G_IO_IN))) {
            cl_log(LOG_WARNING, "Couldn't create source for cmap\n");
            goto out;
        }
        g_source_set_callback(cmap_source, cmap_dispatch_callback, NULL, NULL);
        g_source_attach(cmap_source, NULL);
    }

    if (cmap_get_uint32(cmap_handle, "runtime.config.totem.consensus", &consensus) == CS_OK) {
        cl_log(LOG_NOTICE, "Corosync totem consensus is %d ms", consensus);
        qd.consensus_timeout = consensus;
        /* Take 1s (in milliseconds) off the timeout because we only
            * check once per second */
        if (qd.consensus_timeout > 2000) qd.consensus_timeout -= 1000;
    } else {
        cl_log(LOG_WARNING, "Could not read runtime.config.totem.consensus\n");
        goto out;
    }
    if (cmap_get_uint8(cmap_handle, "quorum.two_node", &two_node_u8) == CS_OK) {
        cl_log(LOG_NOTICE, "Corosync is%s in 2Node-mode", two_node_u8?"":" not");
        two_node = two_node_u8;
    } else {
        cl_log(LOG_NOTICE, "quorum.two_node not present in cmap\n");
    }
    return TRUE;

out:
    if (cmap_source) {
        g_source_destroy(cmap_source);
        cmap_source = NULL;
    }
    if (twonode_track_handle) {
        cmap_track_delete(cmap_handle, twonode_track_handle);
        twonode_track_handle = 0;
    }
    if (consensus_track_handle) {
        cmap_track_delete(cmap_handle, consensus_track_handle);
        consensus_track_handle = 0;
    }
    if (cmap_handle) {
        cmap_finalize(cmap_handle);
        cmap_handle = 0;
    }

    return FALSE;
}

/*
 * quorum_getquorate_thread() is called in a separate thread.
 * It could access the quorum data structure directly but
 * since we can pass it, it is done this way. Currently we do
 * not have any locking on this struct because none is needed;
 * this thread only reads values which are never modified while
 * it is running.
 */
static void
quorum_getquorate_thread(GTask         *task,
                         gpointer       source_object,
                         gpointer       task_data,
                         GCancellable  *cancellable)
{
    int is_quorate;
    int err;
    struct qdata *qd = task_data;

    err = quorum_getquorate(qd->q_handle, &is_quorate);
    if (err != CS_OK) g_task_return_int(task, -1); // Indeterminate result
    else if (is_quorate) g_task_return_int(task, 1); // Quorate
    else g_task_return_int(task, 0); // Not quorate
}

static void
quorum_getquorate_cb(GObject *source,
                     GAsyncResult *result,
                     gpointer user_data)
{
    int is_quorate = -1; // Default to indeterminate result
    GError *error = NULL;

    /* 
     * The callback is called in the main thread same as
     * do_quorum_getquorate() so we do not need a lock to modify
     * the structure. This allows another attempt to start now
     * that the last one finished
     */
    qd.running = 0;
    is_quorate = g_task_propagate_int(G_TASK(result), &error);
    /* 
     * Do not care about the error if we get one; the indeterminate (-1) result
     * is enough. But we do need to free it or we will leak memory.
     */
    if (error) g_error_free(error);
    cl_log(LOG_DEBUG, "is_quorate=%d", is_quorate);
    clock_gettime(CLOCK_MONOTONIC, &qd.ts);
    switch (is_quorate) {
        case 0:
            if (qd.ever_had_quorum) {
                set_servant_health(pcmk_health_noquorum, LOG_CRIT, "Quorum lost");
            } else {
                set_servant_health(pcmk_health_online, LOG_INFO, "We do not have quorum yet");
            }
            break;
        case 1:
            set_servant_health(pcmk_health_online, LOG_INFO, "Node state: online");
            qd.ever_had_quorum = TRUE;
            break;
        default:
            set_servant_health(pcmk_health_transient, LOG_INFO, "Unable to determine quorum status");
    }
}

static void
do_quorum_getquorate(void)
{
    GTask *task;

    /* Do not start another attempt if one is already running */
    if (qd.running) return;
    task = g_task_new (NULL, NULL, quorum_getquorate_cb, NULL);
    g_task_set_task_data(task, &qd, NULL);
    /* Prevent another attempt starting before this one finishes */
    qd.running = 1;
    g_task_run_in_thread(task, quorum_getquorate_thread);
    g_object_unref (task);
}

/* Subtract the ‘struct timeval’ values X and Y,
   storing the result in RESULT.
   Return 1 if the difference is negative, otherwise 0. */

static int
timeval_subtract(struct timespec *result, struct timespec *x, struct timespec *y)
{
    /* Perform the carry for the later subtraction by updating y. */
    if (x->tv_nsec < y->tv_nsec) {
        int nsec = (y->tv_nsec - x->tv_nsec) / 1000000 + 1;
        y->tv_nsec -= 1000000000 * nsec;
        y->tv_sec += nsec;
    }
    if (x->tv_nsec - y->tv_nsec > 1000000000) {
        int nsec = (x->tv_nsec - y->tv_nsec) / 1000000000;
        y->tv_nsec += 1000000000 * nsec;
        y->tv_sec -= nsec;
    }

    /* Compute the time remaining to wait.
        tv_nsec is certainly positive. */
    result->tv_sec = x->tv_sec - y->tv_sec;
    result->tv_nsec = x->tv_nsec - y->tv_nsec;

    /* Return 1 if result is negative. */
    return x->tv_sec < y->tv_sec;
}

#endif /* SUPPORT_COROSYNC */

static gboolean
notify_timer_cb(gpointer data)
{
    cl_log(LOG_DEBUG, "Refreshing %sstate", remote_node?"remote ":"");

    if(remote_node) {
        sbd_remote_check(NULL);
        return TRUE;
    }

    switch (get_cluster_type()) {
        case pcmk_cluster_classic_ais:
            send_cluster_text(crm_class_quorum, NULL, TRUE, NULL, crm_msg_ais);
            break;

        case pcmk_cluster_corosync:
        case pcmk_cluster_cman:
#if SUPPORT_COROSYNC
            /* Initiate asynchronous quorum check */
            do_quorum_getquorate();
            if (qd.ever_had_quorum) {
                struct timespec ts;
                struct timespec diff;
                
                if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0) {
                    /*
                    * If the last successful quorum query state (in seconds) is older than
                    * the consensus timeout (in milliseconds), declare our state to be
                    * indeterminate.
                    */
                    if (timeval_subtract(&diff, &ts, &qd.ts) >= 0) {
                        if (diff.tv_sec * 1000 + diff.tv_nsec / 1000000 > qd.consensus_timeout) {
                                set_servant_health(pcmk_health_transient, LOG_CRIT, "Quorum check timed out, diff %lu, consensus %u", diff.tv_sec, qd.consensus_timeout);
                        }
                    }
                }
            }
#endif /* SUPPORT_COROSYNC */
            notify_parent();
            break;

        default:
            break;
    }
    return TRUE;
}


static void
sbd_membership_connect(void)
{
    bool connected = false;
    int err;

    cl_log(LOG_NOTICE, "Attempting cluster connection");

    cluster.destroy = sbd_membership_destroy;

#if SUPPORT_PLUGIN
    cluster.cpg.cpg_deliver_fn = sbd_plugin_membership_dispatch;
#endif

#if SUPPORT_COROSYNC
    cluster.cpg.cpg_confchg_fn = sbd_cpg_membership_dispatch;
    qd.q_handle = 0;
    qd.ever_had_quorum = FALSE;
#endif

    while(connected == false) {

        enum cluster_type_e stack = get_cluster_type();
        if(get_cluster_type() == pcmk_cluster_unknown) {
            crm_debug("Attempting pacemaker remote connection");
            /* Nothing is up, go looking for the pacemaker remote process */
            if(find_pacemaker_remote() > 0) {
                connected = true;
            }

        } else {
            cl_log(LOG_INFO, "Attempting connection to %s", name_for_cluster_type(stack));

#if SUPPORT_COROSYNC
            if (sbd_crm_setup()) {
#endif

                if(crm_cluster_connect(&cluster)) {
                    connected = true;
                }

#if SUPPORT_COROSYNC
                /* Connect to quorum service so we can use q_handle */
                cl_log(LOG_INFO, "Attempting quorum connection");
                err = quorum_initialize(&qd.q_handle, NULL, &qd.q_type);
                if (err != CS_OK) {
                    cl_log(LOG_ERR, "Cannot initialize QUORUM service: %d\n", err);
                    qd.q_handle = 0;
                    crm_cluster_disconnect(&cluster);
                    connected = false;
                }
            }
#endif
        }

        if(connected == false) {
            cl_log(LOG_INFO, "Failed, retrying in %ds", reconnect_msec / 1000);
            sleep(reconnect_msec / 1000);
        }
    }

    set_servant_health(pcmk_health_transient, LOG_NOTICE, "Connected, waiting for initial membership");
    notify_parent();

    notify_timer_cb(NULL);
}

static void
sbd_membership_destroy(gpointer user_data)
{
    int err;

    cl_log(LOG_WARNING, "Lost connection to %s", name_for_cluster_type(get_cluster_type()));

    set_servant_health(pcmk_health_unclean, LOG_ERR, "Cluster connection terminated");
    notify_parent();

#if SUPPORT_COROSYNC
    if (qd.running) {
        /* Don't have a safe way to wait for the thread to complete
         * here. As a first pass, complain and return if it is, and
         * see if we ever get this */
        cl_log(LOG_CRIT, "sbd_membership_destroy() with running quorum query thread");
        return;
    }
    /* Best effort attempt to disconnect from quorum service */
    cl_log(LOG_INFO, "Attempting quorum disconnection");
    err = quorum_finalize(qd.q_handle);
    if (err != CS_OK) {
        /* The subsequent sbd_membership_connect() will probably
         * fail after this */
        cl_log(LOG_ERR, "Cannot finalize QUORUM service: %d\n", err);
    }
#endif

    /* Attempt to reconnect, the watchdog will take the node down if the problem isn't transient */
    sbd_membership_connect();
}

/*
 * \internal
 * \brief Get process ID and name associated with a /proc directory entry
 *
 * \param[in]  entry    Directory entry (must be result of readdir() on /proc)
 * \param[out] name     If not NULL, a char[64] to hold the process name
 * \param[out] pid      If not NULL, will be set to process ID of entry
 *
 * \return 0 on success, -1 if entry is not for a process or info not found
 *
 * \note This should be called only on Linux systems, as not all systems that
 *       support /proc store process names and IDs in the same way.
 *       Copied from the Pacemaker implementation.
 */
int
sbd_procfs_process_info(struct dirent *entry, char *name, int *pid)
{
    int fd, local_pid;
    FILE *file;
    struct stat statbuf;
    char key[16] = { 0 }, procpath[128] = { 0 };

    /* We're only interested in entries whose name is a PID,
     * so skip anything non-numeric or that is too long.
     *
     * 114 = 128 - strlen("/proc/") - strlen("/status") - 1
     */
    local_pid = atoi(entry->d_name);
    if ((local_pid <= 0) || (strlen(entry->d_name) > 114)) {
        return -1;
    }
    if (pid) {
        *pid = local_pid;
    }

    /* Get this entry's file information */
    strcpy(procpath, "/proc/");
    strcat(procpath, entry->d_name);
    fd = open(procpath, O_RDONLY);
    if (fd < 0 ) {
        return -1;
    }
    if (fstat(fd, &statbuf) < 0) {
        close(fd);
        return -1;
    }
    close(fd);

    /* We're only interested in subdirectories */
    if (!S_ISDIR(statbuf.st_mode)) {
        return -1;
    }

    /* Read the first entry ("Name:") from the process's status file.
     * We could handle the valgrind case if we parsed the cmdline file
     * instead, but that's more of a pain than it's worth.
     */
    if (name != NULL) {
        strcat(procpath, "/status");
        file = fopen(procpath, "r");
        if (!file) {
            return -1;
        }
        if ((fscanf(file, "%15s%63s", key, name) != 2)
            || safe_str_neq(key, "Name:")) {
            fclose(file);
            return -1;
        }
        fclose(file);
    }

    return 0;
}


static gboolean
sbd_remote_check(gpointer user_data)
{
    static int have_proc_pid = 0;

    int running = 0;

    cl_log(LOG_DEBUG, "Checking pacemaker remote connection: %d/%d", have_proc_pid, remoted_pid);
    
    if(have_proc_pid == 0) {
        char proc_path[PATH_MAX], exe_path[PATH_MAX];

        /* check to make sure pid hasn't been reused by another process */
        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", (long unsigned int)getpid());

        have_proc_pid = 1;
        if(readlink(proc_path, exe_path, PATH_MAX - 1) < 0) {
            have_proc_pid = -1;
        }
    }
    
    if (remoted_pid <= 0) {
        set_servant_health(pcmk_health_transient, LOG_WARNING, "No Pacemaker Remote connection");
        goto notify;

    } else if (kill(remoted_pid, 0) < 0 && errno == ESRCH) {
        /* Not running */

    } else if(have_proc_pid == -1) {
        running = 1;
        cl_log(LOG_DEBUG, "Poccess %ld is active", (long)remoted_pid);

    } else {
        int rc = 0;
        char proc_path[PATH_MAX], exe_path[PATH_MAX], expected_path[PATH_MAX];

        /* check to make sure pid hasn't been reused by another process */
        snprintf(proc_path, sizeof(proc_path), "/proc/%lu/exe", (long unsigned int)remoted_pid);

        rc = readlink(proc_path, exe_path, PATH_MAX - 1);
        if (rc < 0) {
            crm_perror(LOG_ERR, "Could not read from %s", proc_path);
            goto done;
        }
        exe_path[rc] = 0;

        rc = snprintf(expected_path, sizeof(proc_path), "%s/pacemaker_remoted", SBINDIR);
        expected_path[rc] = 0;

        if (strcmp(exe_path, expected_path) == 0) {
            cl_log(LOG_DEBUG, "Process %s (%ld) is active",
                   exe_path, (long)remoted_pid);
            running = 1;
        }
    }

  done:
    
    if(running) {
        set_servant_health(pcmk_health_online, LOG_INFO,
                           "Connected to Pacemaker Remote %lu", (long unsigned int)remoted_pid);
    } else {
        set_servant_health(pcmk_health_unclean, LOG_WARNING,
                           "Connection to Pacemaker Remote %lu lost", (long unsigned int)remoted_pid);
    }

  notify:    
    notify_parent();

    if(running == 0) {
        sbd_membership_connect();
    }
    return true;
}

static long unsigned int
find_pacemaker_remote(void)
{
    DIR *dp;
    char entry_name[64];
    struct dirent *entry;

    dp = opendir("/proc");
    if (!dp) {
        /* no proc directory to search through */
        cl_log(LOG_NOTICE, "Can not read /proc directory to track existing components");
        return FALSE;
    }

    while ((entry = readdir(dp)) != NULL) {
        int pid;

        if (sbd_procfs_process_info(entry, entry_name, &pid) < 0) {
            continue;
        }

        /* entry_name is truncated to 16 characters including the nul terminator */
        cl_log(LOG_DEBUG, "Found %s at %u", entry_name, pid);
        if (strcmp(entry_name, "pacemaker_remot") == 0) {
            cl_log(LOG_NOTICE, "Found Pacemaker Remote at PID %u", pid);
            remoted_pid = pid;
            remote_node = true;
            break;
        }
    }

    closedir(dp);

    return remoted_pid;
}

static void
clean_up(int rc)
{
    return;
}

static void
cluster_shutdown(int nsig)
{
    clean_up(0);
}

int
servant_cluster(const char *diskname, int mode, const void* argp)
{
    enum cluster_type_e cluster_stack = get_cluster_type();

    crm_system_name = strdup("sbd:cluster");
    cl_log(LOG_INFO, "Monitoring %s cluster health", name_for_cluster_type(cluster_stack));
    set_proc_title("sbd: watcher: Cluster");

    sbd_membership_connect();

    /* stonith_our_uname = cluster.uname; */
    /* stonith_our_uuid = cluster.uuid; */

    mainloop = g_main_new(FALSE);
    notify_timer = g_timeout_add(timeout_loop * 1000, notify_timer_cb, NULL);

    mainloop_add_signal(SIGTERM, cluster_shutdown);
    mainloop_add_signal(SIGINT, cluster_shutdown);
    
    g_main_run(mainloop);
    g_main_destroy(mainloop);
    
    clean_up(0);
    return 0;                   /* never reached */
}
