/*
 * Copyright (C) Citrix Systems Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/*
 * SBD monitor for xapi health
 */

#include <unistd.h>

#include <crm/cluster.h>
#include <crm/common/mainloop.h>

#include <glib-unix.h>

#include "sbd.h"


#define XAPI_HEALTHCHECKER_PATH "/opt/xensource/libexec/"
#define XAPI_HEALTHCHECKER      "xapi-health-check"
#define HA_ENABLE_PATH		"/var/run/cluster/"
#define HA_ENABLE_FILE		"ha_enabled"

static GMainLoop	*mainloop     = NULL;
static guint		 notify_timer = 0;
static guint		 health_timer = 0;
static guint             health_blocked_timer = 0;
static struct xapi_servant_params *params;

#define		HEALTH_CHECK_PERIOD   	60
#define         DEFAULT_HEALTH_TIMEOUT  120

static GPid       checker_pid = 0;


static void
checker_watch_cb (GPid     pid,
		  gint     status,
		  gpointer user_data);
static gboolean
health_timer_abort_cb(gpointer data);

static gboolean
start_health_checker()
{
	const gchar * const argv[] = {XAPI_HEALTHCHECKER_PATH XAPI_HEALTHCHECKER, NULL };
	GError *error = NULL;

	cl_log(LOG_DEBUG, "Starting checker subprocess, task timeout %d", params->health_check_timeout);
	g_spawn_async(
		NULL,
		(gchar **)argv,
		NULL,
		G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
		NULL,
		NULL,
		&checker_pid,
		&error);

	if (error != NULL) {
		cl_log(LOG_ERR, "Failed to spawn health check %d", error->code);
		set_servant_health(pcmk_health_transient, LOG_INFO, "Node state: xapi health check failed");
		g_error_free(error);
		return FALSE;
	}

	g_child_watch_add(checker_pid, checker_watch_cb, NULL);
	health_blocked_timer = g_timeout_add(params->health_check_timeout * 1000, health_timer_abort_cb, NULL);

	return TRUE;
}

static void
checker_watch_cb (GPid     pid,
		  gint     status,
		  gpointer user_data)
{
	GError *error = NULL;
	cl_log(LOG_INFO, "Checker process watch cb");

	// Remove the abort timer as the process has terminated
	g_source_destroy(g_main_context_find_source_by_id(NULL, health_blocked_timer));

	if (g_spawn_check_exit_status(status, &error)) {
		set_servant_health(pcmk_health_online, LOG_INFO, "Node state: xapi healthy");
	} else {
		cl_log(LOG_ERR, "Xapi health check '%s', %d", error->message, status);
		set_servant_health(pcmk_health_transient, LOG_ERR, "Node state: xapi health check failed");
		notify_parent();
		g_error_free(error);
	}

	g_spawn_close_pid (pid);
	checker_pid = 0;
}

static gboolean
health_timer_abort_cb(gpointer data)
{
	if (checker_pid != 0) {
		cl_log(LOG_CRIT, "Checker process timed out killing");
		kill(checker_pid, SIGKILL);
	}

	return FALSE;
}

static gboolean
health_timer_cb(gpointer data)
{
	gboolean result = TRUE;

	cl_log(LOG_DEBUG, "Health timer cb");

	if (access(HA_ENABLE_PATH HA_ENABLE_FILE, F_OK) != -1) {
		if (checker_pid == 0) {
			result = start_health_checker();
		}
	} else {
		set_servant_health(pcmk_health_online, LOG_INFO, "Node state: xapi healthy");
		checker_pid = 0;
	}

	return result;
}

static gboolean
notify_timer_cb(gpointer data)
{
	cl_log(LOG_DEBUG, "Refreshing state");

	if (access(HA_ENABLE_PATH HA_ENABLE_FILE, F_OK) == -1) {
		set_servant_health(pcmk_health_online, LOG_INFO, "Node state: xapi healthy");
	}

	notify_parent();

	return TRUE;
}

static void
clean_up(int rc)
{
	return;
}

static void
xapi_shutdown(int nsig)
{
	clean_up(0);
}

int
servant_xapi(const char *diskname, int mode, const void* argp)
{
	sigset_t procmask;

	params = (struct xapi_servant_params*)argp;

	cl_log(LOG_INFO, "Monitoring xapi health");
	cl_log(LOG_INFO, "Monitoring xapi health, checker timeout %d",
	       params->health_check_timeout);

	set_proc_title("sbd: watcher: Xapi");

	if (params->health_check_timeout == 0) {
		params->health_check_timeout = DEFAULT_HEALTH_TIMEOUT;
	}

	/* As we use subprocess unblock SIGCHILD */
	sigemptyset(&procmask);
	sigaddset(&procmask, SIGCHLD);
	sigprocmask(SIG_UNBLOCK, &procmask, NULL);

	mainloop = g_main_loop_new(NULL, FALSE);
	notify_timer = g_timeout_add_seconds(timeout_loop, notify_timer_cb, NULL);
	health_timer = g_timeout_add_seconds(HEALTH_CHECK_PERIOD, health_timer_cb, NULL);

	mainloop_add_signal(SIGTERM, xapi_shutdown);
	mainloop_add_signal(SIGINT, xapi_shutdown);

	/* Start in healthy state */
	set_servant_health(pcmk_health_online, LOG_INFO, "Node state: xapi healthy");

	g_main_loop_run(mainloop);
	g_main_loop_unref(mainloop);

	cl_log(LOG_CRIT, "Unexpected termination of Xapi servant main loop");

	clean_up(0);
	return 0;                   /* never reached */
}
